/* -----------------------------------------------------------------------------------------------
	Top Module of masked polynomial basis core
----------------------------------------------------------------------------------------------- */ 

module aes_core_shared_poly(
	input i_clk,
	input i_rst_n,
	input i_encrypt, 		// flag to set encrypt or decrypt mode
	input [0:3][0:3][7:0] i_data,
	input i_data_rdy,
	input i_cbc_enable,
	input [0:3][0:3][7:0] i_sm_1,
	input [0:3][0:3][7:0] i_sm_2,
	input [0:3][0:3][7:0] i_sm_3,

	input [255:0] i_key,
	input [1:0] i_key_rdy,

	input [0:3][0:3][7:0] i_km_1,
	input [0:3][0:3][7:0] i_km_2,

	input [0:15][0:4][3:0] i_rm_bits,
	input [0:3][0:3][7:0] i_iv,

	output o_key_ack,
	output o_data_rdy,
	output [0:3][0:3][7:0] o_data
);


// Sbox signals
logic [0:3][0:3][7:0] sb1; 
logic [0:3][0:3][7:0] sb2;
logic [0:3][0:3][7:0] sb3;
logic [0:3][0:3][7:0] sb4;
logic [0:3][0:3][0:4][3:0] remask_bits;
//logic [0:15][0:2][7:0] sb_o;
logic [0:3][0:3][7:0] sb_o1;
logic [0:3][0:3][7:0] sb_o2;
logic [0:3][0:3][7:0] sb_o3;

logic encrypt; 				

// State reg signals
logic sreg_clk; 
logic [0:2][0:3][0:3][7:0] sreg_in;
logic [0:2][0:3][0:3][7:0] sreg_out;
logic [1:0] slogic_mode;

// Key logic signals
logic [0:3][0:3][7:0] roundkey_s1;
logic [0:3][0:3][7:0] roundkey_s2;
logic [0:3][0:3][7:0] roundkey_s3;

logic [0:3][0:3][7:0] store_roundkey_s1;
logic [0:3][0:3][7:0] store_roundkey_s2;
logic [0:3][0:3][7:0] store_roundkey_s3;

// Key Expansion signals
logic [0:3][0:3][7:0] rkey;
logic [3:0] current_rnd;
logic [1:0] key_rdy_kex;
logic [1:0] key_mode;
logic key_stable_kex;
logic send_key;

// Control FSM signals
logic done;
logic [1:0] flf;

// Round FSM signals
logic [3:0] round;

logic [1:0] mc_mode;
logic [1:0] fsm_mode;
logic key_stable_sync;

logic ctr_enable;
logic rnd_exec;
logic do_rounds;

// CBC signals
logic [0:3][0:3][7:0] cbc_decrypt;
logic [0:3][0:3][7:0] cbc_encrypt;

assign o_data_rdy = flf[1];
assign mc_mode = {encrypt, flf[0] && flf[1]};
assign encrypt = i_encrypt;
assign o_key_ack = key_stable_sync;

assign sreg_in[0] = sb_o1;
assign sreg_in[1] = sb_o2;
assign sreg_in[2] = sb_o3;

clk_div cd(
	.i_clk(i_clk),
	.i_rst_n(i_rst_n),
	.o_sreg_clk(sreg_clk)
);

key_expansion kex(
	.i_clk(i_clk),
	.i_rst_n(i_rst_n),
	.i_key(i_key),
	.i_key_rdy(i_key_rdy),
	.o_send_key(send_key),
	.o_rkey(rkey),
	.o_current_rnd(current_rnd),
	.o_key_rdy(key_rdy_kex),
	.o_key_stable(key_stable_kex)
);

clk_sync_ack csa(
	.i_sreg_clk(sreg_clk), 
	.i_rst_n(i_rst_n),
	.i_key_stable(key_stable_kex), 
	.o_key_stable_sync(key_stable_sync)
);

key_logic_shared klg(
	.i_clk(i_clk),
	.i_rst_n(i_rst_n),
	.i_mode({encrypt, ~send_key}), // encrypt/decrypt + load/store
	.i_saddr(current_rnd),
	.i_laddr(round),
	.i_flf(flf),
	.i_rkey(rkey),
	.i_kmask1(i_km_1),
	.i_kmask2(i_km_2),
	.o_m1_key(roundkey_s1),
	.o_m2_key(roundkey_s2),
	.o_m3_key(roundkey_s3)
);

control_main_fsm cfsm(
	.i_clk(i_clk),
	.i_sreg_clk(sreg_clk),
	.i_rst_n(i_rst_n),
	.i_key_rdy(key_rdy_kex),
	.i_key_stable(key_stable_sync),
	.i_plaintxt_rdy(i_data_rdy),
	.i_lrnd(flf[1]),
//	.i_rnd_exec(rnd_exec),
	.o_key_mode(key_mode), 		// 128/192/256 key 
	.o_do_rounds(do_rounds)
);


sbox128_shared_poly sb(	
	.i_clk(i_clk),
	.i_rst_n(i_rst_n),
	.i_encrypt(encrypt),
	.i_data1(sb1),
	.i_data2(sb2),
	.i_data3(sb3),
	.i_data4(sb4),
	.i_mask(remask_bits),
	.o_data1(sb_o1),
	.o_data2(sb_o2), 
	.o_data3(sb_o3)

);

state_logic share_1(
	.i_clk(sreg_clk),
	.i_rst_n(i_rst_n),
	.i_enable(do_rounds),
	.i_data(sreg_in[0]),
	.i_mc_mode(mc_mode),
	.o_data(sreg_out[0])
);

state_logic share_2(
	.i_clk(sreg_clk),
	.i_rst_n(i_rst_n),
	.i_enable(do_rounds),
	.i_data(sreg_in[1]),
	.i_mc_mode(mc_mode),
	.o_data(sreg_out[1])
);

state_logic share_3(
	.i_clk(sreg_clk),
	.i_rst_n(i_rst_n),
	.i_enable(do_rounds),
	.i_data(sreg_in[2]),
	.i_mc_mode(mc_mode),
	.o_data(sreg_out[2])
);


round_counter rc(
	.i_clk(sreg_clk),
	.i_rst_n(i_rst_n),
	.i_encrypt(encrypt),
	.i_count_enable(do_rounds),
	.i_mode(key_mode),
//	.o_rnd_exec(rnd_exec),
	.o_flf(flf),  // [00 01 11 10] middle, first, last or finished round
	.o_round(round)
);

core_interconnect_shared ci(
	.i_data(i_data),		// plaintext, from outside
	.i_sm_1(i_sm_1),		// state-mask 1
	.i_sm_2(i_sm_2),		// state-mask 2
	.i_sm_3(i_sm_3),		// state-mask 3 (for use in sbox only)
	.i_m1_key(roundkey_s1),		// output from masked key-reg 1
	.i_m2_key(roundkey_s2),		// output from masked key-reg 2
	.i_m3_key(roundkey_s3),		// output from masked key-reg 3
	.i_s1_data(sreg_out[0]),		// output from masked state-reg 1
	.i_s2_data(sreg_out[1]),		// output from masked state-reg 2
	.i_s3_data(sreg_out[2]),		// output from masked state-reg 3
	.i_remask_bits(i_rm_bits),	// bits for sbox-remasking registers
	.i_flf(flf),
	
	.i_cbc_encrypt(cbc_encrypt),
	.i_cbc_decrypt(cbc_decrypt),

	.o_sb1(sb1),
	.o_sb2(sb2),
	.o_sb3(sb3),
	.o_sb4(sb4),
	.o_remask_bits(remask_bits),
	.o_ciphertext(o_data)
);

scbc cbc(
	.i_clk(i_clk),
	.i_rst_n(key_stable_sync),
	.i_encrypt(i_encrypt),
	.i_key_rdy_kex(key_rdy_kex[0] || key_rdy_kex[1]),
	.i_data_rdy(i_data_rdy),
	.i_cbc_enable(i_cbc_enable),
	.i_data_iv(i_iv),
	.i_out_cipher(o_data),
	.i_in_cipher(i_data),
	.i_flf(flf),

	.o_encrypt(cbc_encrypt),
	.o_decrypt(cbc_decrypt)
);

endmodule
